/* Copyright (c) 2009 Ondrej Donek.
   See the file LICENSE.txt for licensing information. */

/**
 * Main object that contains all main functionality of tbNotes.
 * 
 * @type Object
 */
(function() {
	
	/**
	 * This object serves all about storage
	 * 
	 * @private
	 * @link http://developer.mozilla.org/en/Extensions/Thunderbird/HowTos/Common_Thunderbird_Extension_Techniques/Use_SQLite
	 * @type Object
	 */
	this.db = {
		
		/**
		 * Holds database connection
		 * @type Components.interfaces.mozIStorageConnection
		 */
		mDbConnection: null,
		
		/**
		 * Holds object describing our database schema
		 * @type Object
		 */
		mDbSchema: {
			tables: {
				tbn_notes:
					"id				INTEGER PRIMARY KEY, \
					 target_type	TEXT NOT NULL DEFAULT 'none', \
					 target_key		TEXT NOT NULL DEFAULT '', \
					 category_id	INTEGER NOT NULL, \
				     created		TEXT NOT NULL, \
					 updated		TEXT NOT NULL, \
					 title			TEXT NOT NULL, \
					 body			TEXT, \
					 deleted		INTEGER DEFAULT 0",
				tbn_categories:
					"id				INTEGER PRIMARY KEY, \
					 parent_id  	INTEGER, \
					 name       	TEXT NOT NULL, \
					 description	TEXT "
			}
		},
		
		/**
		 * Initialize SQLite database
		 * 
		 * @return void
		 */
		onLoad: function() {
			this.initialized = true;
			this._dbInit();
		},
		
		/**
		 * Initialize database and establish the connection
		 * 
		 * @returns boolean
		 */
		_dbInit: function() {
			var dirService = CC["@mozilla.org/file/directory_service;1"].  
					getService(CI.nsIProperties);  

			var dbFile = dirService.get("ProfD", CI.nsIFile);  
					dbFile.append("tbnotes.sqlite");  

			var dbService = CC["@mozilla.org/storage/service;1"].  
					getService(CI.mozIStorageService);  

			var dbConnection = (!dbFile.exists())
					? this._dbCreate(dbService, dbFile)
					: dbService.openDatabase(dbFile);

			this.mDbConnection = dbConnection;
		},
		
		/**
		 * Create database
		 * 
		 * @return {Components.interfaces.mozIStorageConnection}
		 */
		_dbCreate: function(aDbService, aDbFile) {
			var dbConnection = aDbService.openDatabase(aDbFile);
			this._dbCreateTables(dbConnection);
			
			return dbConnection;
		},
		
		/**
		 * Creates tables
		 * 
		 * @return void
		 */
		_dbCreateTables: function(aDbConnection) {
			for(var name in this.mDbSchema.tables) {
				aDbConnection.createTable(name, this.mDbSchema.tables[name]);
			}
		}
	};
		
	// ======================================================================
	// Event handlers
	
	/**
	 * Called when Thunderbird is loaded
	 * 
	 * @return void
	 */
	this.onLoad = function() {
		// Initialize tbNotes SQLite storage
		extensions.tbnotes.db.onLoad();
		
		// Add new col to the main Thunderbird's messages tree
		var obSrv = CC["@mozilla.org/observer-service;1"].
				getService(CI.nsIObserverService);
		
		obSrv.addObserver(gTbNotesCreateDbObserver, "MsgCreateDBView", false);
		
		// Add event listener for mail context menu
		var context = document.getElementById("mailContext");
		if(context) {
			context.addEventListener("popupshowing", 
					extensions.tbnotes.onMailContextPopupShowing, 
					false);
		}
	};
	
	/**
	 * Fired when Thunderbird going to be quit
	 * 
	 * @return void
	 */
	this.onUnload = function() {
		// TODO Remove all listeners and observers!
	};
	
	/**
	 * Fired when mail context popupmenu is showing.
	 * 
	 * @param aEvent {Components.interfaces.nsIDOMEvent}
	 * @return boolean
	 */
	this.onMailContextPopupShowing = function(aEvent) {
		var enable_cmd = false;
		var menuitem1 = document.getElementById("tbn-editnote-menuitem");
		var menuitem2 = document.getElementById("tbn-removenote-menuitem");
		if(!menuitem1 || !menuitem2) {
			return;
		}
		
		// Check if any mail message is selected
		var selMsg = extensions.tbnotes.getSelectedMessageDescription();
		if(selMsg != null) {
			if(extensions.tbnotes.checkNoteForMessageKey(selMsg.key) === true) {
				enable_cmd = true;
			}
		}
		
		// Enable/disable "edit note" command
		if(enable_cmd === true) {
			if(menuitem1.hasAttribute("disabled")) {
				menuitem1.removeAttribute("disabled");
			}
			if(menuitem2.hasAttribute("disabled")) {
				menuitem2.removeAttribute("disabled");
			}
		} else {
			menuitem1.setAttribute("disabled", "true");
			menuitem2.setAttribute("disabled", "true");
		}
		
		return true;
	};
	
	/**
	 * Performs command 
	 * 
	 * @param aCmdName {string}
	 * @return void
	 */
	this.doCommand = function(aCmdName) {
		switch(aCmdName) {
			// adding new note
			case "add_note":
				extensions.tbnotes.showAddNoteDialog();
				break;
			
			// edit selected note or edit note which is attached 
			// to the selected email message
			case "edit_note":
				var selMsg = extensions.tbnotes.getSelectedMessageDescription();
				var note = (selMsg != null) 
						? extensions.tbnotes.getNoteByMessageKey(selMsg.key) 
						: null;
				
				if(note != null) {
					extensions.tbnotes.showEditNoteDialog(note, selMsg);
				} else {
					// TODO Use stringbundles and prompts service!
					alert("There is no note to edit!");
				}
				break;
			
			// removes note from the selected note
			case "remove_note":
				var selMsg = extensions.tbnotes.getSelectedMessageDescription();
				var note = (selMsg != null) 
						? extensions.tbnotes.getNoteByMessageKey(selMsg.key) 
						: null;
				
				if(note != null) {
					extensions.tbnotes.deleteNote(note);
				} else {
					// TODO Use stringbundles and prompts service!
					alert("There is no note to remove!");
				}
				break;
			
			default:
				CU.reportError("extensions.tbnotes.doCommand: Unknown command ['" + aCmdName + "']!");
				break;
		}
	};
	
	// ======================================================================
	// Normal methods
	
	/**
	 * Returns simple object describing currently selected mail message or NULL.
	 * 
	 * @return object|NULL
	 */
	this.getSelectedMessageDescription = function() {
		if(gFolderDisplay.selectedMessage != null) {
		//if(gMessageDisplay.displayedMessage != null) {
			return {
				key: gFolderDisplay.selectedMessage.messageKey,
				subject: gFolderDisplay.selectedMessage.subject
			};
		}
		
		return null;
	};
	
	/**
	 * Returns TRUE if an email message with given key has any note attached.
	 * 
	 * @param aKey {string}
	 * @return boolean
	 */
	this.checkNoteForMessageKey = function(aKey) {
		var query = "SELECT " + 
		            "  id " + 
					"FROM tbn_notes " + 
					"WHERE " + 
					"  deleted = 0 AND " + 
		            "  target_type = 'message' AND " + 
					"  target_key = :target_key " + 
					"LIMIT 1 ";
		var stmt = extensions.tbnotes.db.mDbConnection.createStatement(query);
		stmt.params.target_key = aKey;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) return true;
		}
		
		return false;
	};
	
	/**
	 * Returns note which is attached to an email message with given key. 
	 * Otherwise returns NULL.
	 * 
	 * @param aKey {string}
	 * @return object|NULL
	 */
	this.getNoteByMessageKey = function(aKey) {
		var query = "SELECT * " + 
		            "FROM tbn_notes " + 
					"WHERE " + 
					"  target_type = 'message' AND " + 
					"  target_key = :target_key AND " + 
					"  deleted = 0 " + 
					"LIMIT 1 ";
		var stmt = extensions.tbnotes.db.mDbConnection.createStatement(query);
		stmt.params.target_key = aKey;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
				var note = {
					id: 			stmt.row.id,
					target_type: 	"message",
					target_key:     aKey,
					category_id: 	stmt.row.category_id,
					created: 		stmt.row.created,
					updated: 		stmt.row.updated,
					title: 			stmt.row.title,
					body: 			stmt.row.body,
					deleted:        ((stmt.row.deleted == 1) ? true : false)
				};
				
				return note;
			}
		}
		
		return null;
	};
	
	/**
	 * Show dialog for adding new note
	 * 
	 * @return void
	 */
	this.showAddNoteDialog = function() {
		// Check if any message is currently selected
		var selMsg = extensions.tbnotes.getSelectedMessageDescription();
		
		// Open dialog self
		var params = { inn: { msg: selMsg }, out: null };
		window.openDialog(TBNOTES_NOTEDLG_URI, "", TBNOTES_NOTEDLG_FTRS, params).focus();
		
		if(params.out) {
			extensions.tbnotes.createNote(params.out);
		}
	};
	
	/** 
	 * Shows dialog for editing note
	 * 
	 * @param aNote {object}
	 * @type aSelMsg {Object} Selected message. Used because of title.
	 * @returned void
	 */
	this.showEditNoteDialog = function(aNote, aSelMsg) {
		var params = { inn: { msg: aSelMsg, note: aNote }, out: null };
		window.openDialog(TBNOTES_NOTEDLG_URI, "", TBNOTES_NOTEDLG_FTRS, params).focus();
		
		if(params.out) {
			extensions.tbnotes.updateNote(params.out);
		}
	};
	
	/**
	 * Adds new note record into the database
	 * 
	 * @param aNote {Object}
	 * @return void
	 */
	this.createNote = function(aNote) {
		try {
			var query = 
					"INSERT INTO tbn_notes " + 
					"  (target_type, target_key, category_id, created, updated, title, body, deleted) " + 
					"VALUES " + 
					"  ( :target_type , :target_key , 0, :created , :updated, :title , :body , 0); ";
			
			var stmt = extensions.tbnotes.db.mDbConnection.createStatement(query);
			stmt.params.target_type = aNote.target;
			stmt.params.target_key = aNote.targetKey;
			stmt.params.created = getCurrentDatetime();
			stmt.params.updated = getCurrentDatetime();
			stmt.params.title = aNote.title;
			stmt.params.body = aNote.body;
			stmt.execute();
		} catch(e) {
			CU.reportError(e);
			
			with(extensions.tbnotes) {
				alert("Database error: [" + db.mDbConnection.lastError + "]: " + 
				      db.mDbConnection.lastErrorString);
			}
		}
	};
	
	/**
	 * Updates note's record in the database
	 * 
	 * @param aNote {Object}
	 * @return void
	 */
	this.updateNote = function(aNote) {
		try {
			// FIXME This query doesn't working!
			var query = 
					"UPDATE tbn_notes SET " + 
					"  updated = :updated , " + 
					"  title = :title ,  " + 
					"  body = :body " + 
					"WHERE " + 
					"  id = :note_id ";
			
			var stmt = extensions.tbnotes.db.mDbConnection.createStatement(query);
			stmt.params.updated = getCurrentDatetime();
			stmt.params.title = aNote.title;
			stmt.params.body = aNote.body;
			stmt.params.note_id = aNote.id;
			stmt.execute();
		} catch(e) {
			CU.reportError(e);
			
			with(extensions.tbnotes) {
				alert("Database error: [" + db.mDbConnection.lastError + "]: " + 
				      db.mDbConnection.lastErrorString);
			}
		}
	};
	
	/**
	 * Mark note as deleted
	 * 
	 * @param aNote {Object}
	 * @return void
	 */
	this.deleteNote = function(aNote) {
		var pref = TBNOTES_PREFS_KEY + "dont_show_remove_confirm";
		
		// Check if user want to see confirmation dialog
		if(TBNOTES_PREFS.prefHasUserValue(pref) === true) {
			if(TBNOTES_PREFS.getBoolPref(pref) === true) {
				return true;
			}
		}
		
		// Show confirmation if user really want to close unsaved note
		var prompts = CC["@mozilla.org/embedcomp/prompt-service;1"].
				getService(CI.nsIPromptService);
		
		// TODO Use stringbundles!
		var title = "Note dialog";
		var desc = "Do you really want to remove note '" + aNote.title + "'?";
		var question = "Do not ask me again.";
		var check = { value: false };
		var res = prompts.confirmCheck(window, title, desc, question, check);
		
		TBNOTES_PREFS.setBoolPref(pref, check.value);
		
		if(!res) {
			return;
		}
		
		try {
			var query = "UPDATE tbn_notes SET deleted = 1 WHERE id = :note_id ";
			var stmt = extensions.tbnotes.db.mDbConnection.createStatement(query);
			stmt.params.note_id = aNote.id;
			stmt.execute();
		} catch(e) {
			CU.reportError(e);
			
			with(extensions.tbnotes) {
				alert("Database error: [" + db.mDbConnection.lastError + "]: " + 
				      db.mDbConnection.lastErrorString);
			}
		}
	};
	
}).apply(extensions.tbnotes);
